from typing import Literal

import anyio
import anyio.abc
from aiodocker import Docker, containers
from anyio.streams.memory import MemoryObjectReceiveStream, MemoryObjectSendStream
from loguru import logger
from pydantic import BaseModel, Field


class DockerServerParameters(BaseModel):
    """Configuration for Docker transport."""

    image: str = Field(description="Image of the docker container")
    args: list[str] = Field(
        default_factory=list,
        description="Command line arguments for the docker container",
    )
    env: dict[str, str] = Field(
        default_factory=dict,
        description="Environment variables for the docker container",
    )


class DockerTransport:
    state: Literal["constructed", "initialized", "started", "stopped"] = "constructed"

    def __init__(self, config: DockerServerParameters) -> None:
        self.config = config
        self.docker: Docker | None = None
        self.container: containers.DockerContainer | None = None
        self.attach_result: any = None

        self._reader_send: MemoryObjectSendStream[str]
        self._reader_recv: MemoryObjectReceiveStream[str]

        self._writer_send: MemoryObjectSendStream[str]
        self._writer_recv: MemoryObjectReceiveStream[str]

        self._task_group: anyio.abc.TaskGroup | None = None

    async def init(self) -> None:
        self.docker = Docker()
        await self.docker.images.pull(self.config.image)

        self._reader_send, self._reader_recv = anyio.create_memory_object_stream()
        self._writer_send, self._writer_recv = anyio.create_memory_object_stream()

        self.state = "initialized"
        logger.debug("DockerTransport initialized")

    async def start(self) -> None:
        if not self.docker:
            raise RuntimeError("Docker not initialized")

        self.container = await self.docker.containers.create(
            {
                "Image": self.config.image,
                "Cmd": self.config.args,
                "OpenStdin": True,
                "AttachStdin": True,
                "AttachStdout": True,
                "AttachStderr": True,
                "Tty": False,
                "Env": [f"{k}={v}" for k, v in self.config.env.items()],
                "HostConfig": {"AutoRemove": True},
            }
        )

        await self.container.start()
        logger.debug(f"Started container {self.container.id}")

        self.attach_result = self.container.attach(stdin=True, stdout=True, stderr=True)

        async def read_stdout():
            assert self.attach_result is not None

            buffer = ""
            async with self._reader_send:
                while True:
                    try:
                        msg = await self.attach_result.read_out()
                        if msg is None:
                            logger.debug("Docker stream closed (EOF)")
                            break
                        if msg.data is None:
                            continue

                        if isinstance(msg.data, (bytes, bytearray)):
                            chunk = msg.data.decode("utf-8", errors="replace")
                        else:
                            chunk = str(msg.data)

                        buffer += chunk
                        while "\n" in buffer:
                            line, buffer = buffer.split("\n", 1)
                            await self._reader_send.send(line)
                    except Exception as e:
                        logger.error(f"Error reading container output: {e}")
                        break
            logger.debug("read_stdout task exited")

        async def write_stdin():
            assert self.attach_result is not None
            async with self._writer_recv:
                async for msg in self._writer_recv:
                    try:
                        await self.attach_result.write_in(msg.encode("utf-8") + b"\n")
                    except Exception as e:
                        logger.warning(f"Failed to write to container stdin: {e}")
            logger.debug("write_stdin task exited")

        self._task_group = anyio.create_task_group()
        await self._task_group.__aenter__()
        self._task_group.start_soon(read_stdout)
        self._task_group.start_soon(write_stdin)

        self.state = "started"

    async def stop(self) -> None:
        logger.debug("Stopping DockerTransport...")

        if self.attach_result:
            try:
                await self.attach_result.close()
            except Exception:
                pass
            self.attach_result = None

        if self.container:
            try:
                await self.container.stop()
            except Exception:
                pass
            try:
                await self.container.delete()
            except Exception:
                pass
            self.container = None

        # Close input stream to stop write_stdin task
        try:
            await self._writer_send.aclose()
        except Exception:
            pass

        if self._task_group:
            await self._task_group.__aexit__(None, None, None)
            self._task_group = None

        if self.docker:
            await self.docker.close()
            self.docker = None

        self.state = "stopped"
        logger.debug("DockerTransport stopped and cleaned up")


    async def send(self, message: str) -> None:
        if self.state != "started":
            raise RuntimeError("Transport not started")
        await self._writer_send.send(message)

    async def receive(self) -> str:
        if self.state != "started":
            raise RuntimeError("Transport not started")
        return await self._reader_recv.receive()



# Demo of how to use easymcp to perform RAG via mcp resources. Pair with a good filesystem mcp server to use your local documents dynamically.
# to run this example you need to install meilisearch python sdk, and run a meilisearch server on localhost:7700
# uv add easymcp meilisearch

import asyncio
import meilisearch.client 
from easymcp.client.ClientManager import ClientManager
from easymcp.client.transports.stdio import StdioServerParameters
from mcp.types import TextResourceContents
from itertools import count

# disable logging
from loguru import logger
logger.disable("easymcp")

# create a Meilisearch client
client = meilisearch.Client('http://127.0.0.1:7700')

# create an index
index = client.index('mcp-resources')

# create a ClientManager
mgr = ClientManager()

# define the servers
servers = {
    "timeserver": StdioServerParameters(
        command="uvx",
        args=["mcp-timeserver"]
    ),
    "llms-txt": StdioServerParameters(
        command="uvx",
        args=["mcp-llms-txt"]
    ),
}

async def main():
    # initialize the client manager and start the servers
    await mgr.init(servers)
    
    print("reading resources") # list resources
    resources = await mgr.list_resources()

    documents = []

    counter = count()

    print("indexing resources")
    for resource in resources:
        documents.append({
            "id": next(counter),
            "uri": str(resource.uri),
            "name": resource.name,
            "description": resource.description,
            "mime_type": resource.mimeType,
        })

    task = index.add_documents(documents) # add documents to the index
    index.wait_for_task(task.task_uid)

    print("indexing done")

    QUERY = "london"

    results = index.search(QUERY) # search for a resource
    hits = results.get("hits", [])

    print()

    for hit in hits:
        print(f"Result: {hit['name']}: {hit['uri']}")
        result = await mgr.read_resource(hit["uri"]) # resolve the resource uri

        for c in result.contents:
            if isinstance(c, TextResourceContents):
                print(c.text)

        print()

    print("done")

    await asyncio.Future()

asyncio.run(main())


import asyncio
from easymcp.client.sessions.mcp import MCPClientSession
from easymcp.client.transports.stdio import StdioTransport, StdioServerParameters


async def main():
    args = StdioServerParameters(command="uvx", args=["mcp-wolfram-alpha"], env={"WOLFRAM_API_KEY": "DEMO"})
    transport = StdioTransport(args)

    client_session = MCPClientSession(transport)
    await client_session.init()

    result = await client_session.start()
    print(f"{result=}")

    # await asyncio.sleep(10)

    prompts = await client_session.list_prompts()
    print(f"{prompts=}")

    resolved = await client_session.read_prompt("wa", {"query": "pi"})
    print(f"{resolved=}")

    # await client_session.stop()
    await asyncio.Future()


asyncio.run(main())



from typing import Protocol, runtime_checkable, Awaitable, Callable

from mcp import types

# === Core Capability Protocols ===

@runtime_checkable
class ToolsCompatible(Protocol):
    async def list_tools(self, force: bool = False) -> types.ListToolsResult: ...
    async def call_tool(self, tool_name: str, args: dict) -> types.CallToolResult: ...


@runtime_checkable
class ResourcesCompatible(Protocol):
    async def list_resources(self, force: bool = False) -> types.ListResourcesResult: ...
    async def read_resource(self, resource_name: str) -> types.ReadResourceResult: ...


@runtime_checkable
class PromptsCompatible(Protocol):
    async def list_prompts(self, force: bool = False) -> types.ListPromptsResult: ...
    async def read_prompt(self, prompt_name: str, args: dict) -> types.GetPromptResult: ...


# === Session Lifecycle Protocols ===

@runtime_checkable
class InitializerProtocol(Protocol):
    async def init(self) -> None: ...


@runtime_checkable
class LifeSpanProtocol(InitializerProtocol, Protocol):
    async def start(self) -> types.InitializeResult: ...
    async def stop(self) -> None: ...


# === Reactive / Push-capable Protocols ===

@runtime_checkable
class PushingToolsCompatible(ToolsCompatible, Protocol):
    async def register_tools_changed_callback(
        self, callback: Callable[[], Awaitable[None]]
    ) -> None: ...


@runtime_checkable
class PushingPromptsCompatible(PromptsCompatible, Protocol):
    async def register_prompts_changed_callback(
        self, callback: Callable[[], Awaitable[None]]
    ) -> None: ...


@runtime_checkable
class PushingResourcesCompatible(ResourcesCompatible, Protocol):
    async def register_resources_changed_callback(
        self, callback: Callable[[], Awaitable[None]]
    ) -> None: ...


@runtime_checkable
class PushingRootsCompatible(Protocol):
    async def register_roots_callback(
        self, callback: Callable[[types.ListRootsRequest], Awaitable[types.ListRootsResult]]
    ) -> None: ...


@runtime_checkable
class PushingSamplingCompatible(Protocol):
    async def register_sampling_callback(
        self, callback: Callable[[types.CreateMessageRequest], Awaitable[types.CreateMessageResult]]
    ) -> None: ...

# === Top-Level Session Protocols ===

@runtime_checkable
class BaseSessionProtocol(InitializerProtocol, Protocol): ...


import asyncio

from easymcp.client.ClientManager import ClientManager
from easymcp.client.transports.stdio import StdioServerParameters

mgr = ClientManager()

searxng = StdioServerParameters(
    command="uvx",
    args=["mcp-searxng"],
)

timeserver = StdioServerParameters(
    command="uvx",
    args=["mcp-timeserver"],
)

servers = {
    "searxng": searxng,
    "timeserver": timeserver,
}

async def main():
    print("Listing servers")
    await mgr.init(servers=servers)
    print(mgr.list_servers())

    # print("Removing searxng")
    # await mgr.remove_server("searxng")
    # print(mgr.list_servers())

    # print("Adding searxng")
    # await mgr.add_server("searxng", searxng)
    # print(mgr.list_servers())

    print("Listing tools")
    print(await mgr.list_tools())

    print("Calling tool")
    print(await mgr.call_tool("timeserver.get-current-time", {}))

    print("Listing resources")
    print(await mgr.list_resources())

    print("Reading resource")
    # await mgr.sessions["timeserver"].read_resource("datetime://Africa/Algiers/now")
    print(await mgr.read_resource("mcp-timeserver+datetime://Africa/Algiers/now"))


    await asyncio.Future()

asyncio.run(main())


from asyncio import Future, Queue

from mcp import types

from loguru import logger


class RequestMap:
    """RequestMap class"""

    requests: dict[str, Future[types.JSONRPCResponse | None]]

    outgoing_messages: Queue[types.JSONRPCMessage]

    def __init__(self, outgoing_messages: Queue[types.JSONRPCMessage]):
        self.outgoing_messages = outgoing_messages
        self.requests = {}

    def send_request(self, message: types.JSONRPCRequest):
        """send a request"""

        logger.debug(f"Sending request: {message}")

        future = Future[types.JSONRPCResponse | None]()

        self.requests[str(message.id)] = future

        wrapped_message = types.JSONRPCMessage(message)
        self.outgoing_messages.put_nowait(wrapped_message)
        return future

    def resolve_request(self, message: types.JSONRPCResponse):
        """resolve a request"""

        logger.debug(f"Resolving request: {message}")

        request_id = message.id
        future = self.requests.pop(str(request_id), None)
        if future is not None:
            future.set_result(message)

    def resolve_error(self, message: types.JSONRPCError):
        """resolve an error"""

        logger.debug(f"Resolving error: {message}")

        request_id = message.id
        future = self.requests.pop(str(request_id), None)
        if future is not None:
            future.set_result(None)


from typing import TypeAlias

from easymcp.client.sessions.GenericSession import BaseSessionProtocol
from easymcp.client.sessions.mcp import MCPClientSession
from easymcp.client.transports.stdio import StdioTransport, StdioServerParameters
from easymcp.client.transports.docker import DockerTransport, DockerServerParameters
from easymcp.client.transports.sse import SseTransport, SseServerParameters


transportTypes: TypeAlias = StdioServerParameters | DockerServerParameters | SseServerParameters

def make_transport(arguments: transportTypes) -> BaseSessionProtocol:

    if isinstance(arguments, StdioServerParameters):
        return MCPClientSession(StdioTransport(arguments))

    if isinstance(arguments, DockerServerParameters):
        return MCPClientSession(DockerTransport(arguments))
    
    if isinstance(arguments, SseServerParameters):
        return MCPClientSession(SseTransport(arguments))
    
    raise ValueError(f"Unknown transport type: {type(arguments)}")



import asyncio
import time
from easymcp.client.sessions.mcp import MCPClientSession
from easymcp.client.transports.stdio import StdioTransport, StdioServerParameters

class AsyncTimer:
    def __enter__(self):
        self.start = time.perf_counter()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.end = time.perf_counter()
        self.duration = self.end - self.start
        print(f"Elapsed time: {self.duration:.6f} seconds")

async def main():
    args = StdioServerParameters(command="uvx", args=["mcp-wolfram-alpha"], env={"WOLFRAM_API_KEY": "DEMO"})
    transport = StdioTransport(args)

    client_session = MCPClientSession(transport)
    await client_session.init()

    result = await client_session.start()
    print(f"{result=}")

    # await asyncio.sleep(10)

    with AsyncTimer():
        prompts = await client_session.list_prompts()
        print(f"{prompts=}")

    with AsyncTimer():
        prompts = await client_session.list_prompts()
        print(f"{prompts=}")

    with AsyncTimer():
        prompts = await client_session.list_prompts(force=True)
        print(f"{prompts=}")

    # await client_session.stop()
    await asyncio.Future()


asyncio.run(main())



from easymcp.client.utils import format_server_name

assert format_server_name("test") == "test"
assert format_server_name("test-test") == "test-test"
assert format_server_name("test_test") == "test-test"
assert format_server_name("test.test") == "test-test"
assert format_server_name("test.test.test") == "test-test-test"
assert format_server_name("test_test_test") == "test-test-test"


from asyncio import Queue, Task
from inspect import iscoroutinefunction
import json
from typing import Awaitable, Callable

from loguru import logger

from easymcp.client.iobuffers import reader, writer
from easymcp.client.requestmap import RequestMap
from easymcp.client.transports.generic import TransportProtocol

from easymcp.client.utils import CreateJsonRPCRequest
from mcp import types

from easymcp.client.sessions.GenericSession import BaseSessionProtocol

class MCPClientSession(BaseSessionProtocol):
    """ClientSession class"""

    incoming_messages: Queue[types.JSONRPCMessage]
    outgoing_messages: Queue[types.JSONRPCMessage]

    reader_task: Task[None]
    writer_task: Task[None]

    request_map: RequestMap

    roots_callback: Callable[[types.ListRootsRequest], Awaitable[types.ListRootsResult]] | None = None
    sampling_callback: Callable[[types.CreateMessageRequest], Awaitable[types.CreateMessageResult]] | None = None

    tools_changed_callback: Callable[[], Awaitable[None]] | None = None
    prompts_changed_callback: Callable[[], Awaitable[None]] | None = None
    resources_changed_callback: Callable[[], Awaitable[None]] | None = None

    _tools: types.ListToolsResult | None = None
    _prompts: types.ListPromptsResult | None = None
    _resources: types.ListResourcesResult | None = None

    def __init__(self, transport: TransportProtocol):
        self.transport = transport

        # define message queues
        self.incoming_messages = Queue()
        self.outgoing_messages = Queue()

        self._tools = None

    async def init(self):
        """initialize the client session"""
        await self.transport.init()
        self.request_map = RequestMap(self.outgoing_messages)

    @staticmethod
    def _validate_async_callback(callback: Callable, name: str):
        assert callable(callback), f"{name} must be callable"
        assert iscoroutinefunction(callback), f"{name} must be an async function"

    async def register_roots_callback(self, callback: Callable[[types.ListRootsRequest], Awaitable[types.ListRootsResult]]):
        """register a callback for roots"""
        self._validate_async_callback(callback, "roots_callback")
        self.roots_callback = callback

    async def register_sampling_callback(self, callback: Callable[[types.CreateMessageRequest], Awaitable[types.CreateMessageResult]]):
        """register a callback for sampling"""
        self._validate_async_callback(callback, "sampling_callback")
        self.sampling_callback = callback

    async def register_tools_changed_callback(self, callback: Callable[[], Awaitable[None]]):
        """register a callback for tools changed"""
        self._validate_async_callback(callback, "tools_changed_callback")
        self.tools_changed_callback = callback

    async def register_prompts_changed_callback(self, callback: Callable[[], Awaitable[None]]):
        """register a callback for prompts changed"""
        self._validate_async_callback(callback, "prompts_changed_callback")
        self.prompts_changed_callback = callback

    async def register_resources_changed_callback(self, callback: Callable[[], Awaitable[None]]):
        """register a callback for resources changed"""
        self._validate_async_callback(callback, "resources_changed_callback")
        self.resources_changed_callback = callback

    def _start_reading_messages(self):
        async def __start_reading_messages():
            while self.transport.state == "started":
                message = await self.incoming_messages.get()
                if message is None:
                    continue

                # handle responses
                if isinstance(message.root, types.JSONRPCResponse):
                    self.request_map.resolve_request(message.root)

                # handle notifications
                elif isinstance(message.root, types.JSONRPCNotification):
                    if message.root.params is None:
                        logger.error(f"Received notification with no params: {message}")
                        continue

                    notification = types.ServerNotification.model_validate(message.root)

                    await self.handle_notification(notification)

                # handle requests
                elif isinstance(message.root, types.JSONRPCRequest):
                    if message.root.params is None:
                        logger.error(f"Received request with no params: {message}")
                        continue

                    request = types.ServerRequest.model_validate(message.root)

                    response = await self.handle_request(request)
                    if response is not None:
                        response_message = types.JSONRPCResponse(
                            jsonrpc="2.0",
                            id=message.root.id,
                            result=response,
                        )
                        await self.transport.send(response_message.model_dump_json())

                # handle errors
                elif isinstance(message.root, types.JSONRPCError):
                    data = message.model_dump()
                    try:
                        data["error"]["message"] = json.loads(
                            data.get("error").get("message", "{}")
                        )
                        data = json.dumps(data, indent=4)
                        logger.error(f"Received error:\n{data}")
                    except Exception:
                        pass

                    if message.root.id is not None:
                        self.request_map.resolve_error(message.root)

                else:
                    logger.error(f"Unknown message type: {message.root}")

        Task(__start_reading_messages())

    async def start(self) -> types.InitializeResult:
        """start the client session"""
        await self.transport.start()
        self.reader_task = await reader(self.transport, self.incoming_messages)
        self.writer_task = await writer(self.transport, self.outgoing_messages)

        self._start_reading_messages()

        sampling = types.SamplingCapability()
        roots = types.RootsCapability(listChanged=True)

        # send initialize request
        request = types.ClientRequest(
            types.InitializeRequest(
                method="initialize",
                params=types.InitializeRequestParams(
                    protocolVersion=types.LATEST_PROTOCOL_VERSION,
                    capabilities=types.ClientCapabilities(
                        sampling=sampling,
                        experimental={},
                        roots=roots,
                    ),
                    clientInfo=types.Implementation(name="easymcp", version="0.1.0"),
                ),
            )
        )

        response = await self.request_map.send_request(CreateJsonRPCRequest(request))  # type: ignore
        if response is None:
            raise RuntimeError("Failed to initialize client session")

        # send initialized notification
        notification = types.ClientNotification(
            types.InitializedNotification(method="notifications/initialized")
        )

        self.outgoing_messages.put_nowait(
            types.JSONRPCNotification(
                jsonrpc="2.0",
                **notification.model_dump(
                    by_alias=True, mode="json", exclude_none=True
                ),
            )  # type: ignore
        )

        result = types.InitializeResult.model_validate(response.result)
        return result

    async def stop(self):
        """stop the client session"""
        await self.transport.stop()

    async def list_tools(self, force: bool = False):
        """list available tools"""

        if not force and self._tools is not None:
            return self._tools.model_copy(deep=True)

        request = types.ClientRequest(
            types.ListToolsRequest(
                method="tools/list",
            )
        )

        response = await self.request_map.send_request(CreateJsonRPCRequest(request))

        if response is None:
            result = types.ListToolsResult(tools=[])
        else:
            result = types.ListToolsResult.model_validate(response.result)

        self._tools = result.model_copy(deep=True)

        return result

    async def call_tool(self, tool_name: str, args: dict):
        """call a tool"""
        request = types.ClientRequest(
            types.CallToolRequest(
                method="tools/call",
                params=types.CallToolRequestParams(
                    name=tool_name,
                    arguments=args,
                ),
            )
        )

        response = await self.request_map.send_request(CreateJsonRPCRequest(request))
        
        if response is None:
            raise RuntimeError("Failed to call tool")

        result = types.CallToolResult.model_validate(response.result)

        return result

    async def list_resources(self, force: bool = False):
        """list available resources"""

        if not force and self._resources is not None:
            return self._resources.model_copy(deep=True)

        request = types.ClientRequest(
            types.ListResourcesRequest(
                method="resources/list",
            )
        )

        response = await self.request_map.send_request(CreateJsonRPCRequest(request))

        if response is None:
            result = types.ListResourcesResult(resources=[])
        else:
            result = types.ListResourcesResult.model_validate(response.result)

        self._resources = result.model_copy(deep=True)

        return result

    async def read_resource(self, resource_name: str):
        """read a resource"""

        request = types.ClientRequest(
            types.ReadResourceRequest(
                method="resources/read",
                params=types.ReadResourceRequestParams(
                    # TODO: validate uri
                    uri=resource_name,  # type: ignore
                ),
            )
        )

        response = await self.request_map.send_request(CreateJsonRPCRequest(request))
        
        if response is None:
            raise RuntimeError("Failed to read resource")
        
        result = types.ReadResourceResult.model_validate(response.result)

        return result

    async def list_prompts(self, force: bool = False):
        """list available prompts"""

        if not force and self._prompts is not None:
            return self._prompts.model_copy(deep=True)

        request = types.ClientRequest(
            types.ListPromptsRequest(
                method="prompts/list",
            )
        )

        response = await self.request_map.send_request(CreateJsonRPCRequest(request))

        if response is None:
            result = types.ListPromptsResult(prompts=[])
        else:
            result = types.ListPromptsResult.model_validate(response.result)

        self._prompts = result.model_copy(deep=True)

        return result
    
    async def read_prompt(self, prompt_name: str, args: dict):
        """read a prompt"""

        request = types.ClientRequest(
            types.GetPromptRequest(
                method="prompts/get",
                params=types.GetPromptRequestParams(
                    name=prompt_name,
                    arguments=args,
                ),
            )
        )

        response = await self.request_map.send_request(CreateJsonRPCRequest(request))
        result = types.GetPromptResult.model_validate(response.result)

        return result
    
    async def handle_notification(self, notification: types.ServerNotification):
        """handle a notification"""

        logger.debug(f"Handling notification: {notification}")

        if isinstance(notification.root, types.ToolListChangedNotification):
            self._tools = None
            logger.debug("cleared tools cache")
            if self.tools_changed_callback is not None:
                await self.tools_changed_callback()

        elif isinstance(notification.root, types.PromptListChangedNotification):
            self._prompts = None
            logger.debug("cleared prompts cache")
            if self.prompts_changed_callback is not None:
                await self.prompts_changed_callback()

        elif isinstance(notification.root, types.ResourceListChangedNotification):
            self._resources = None
            logger.debug("cleared resources cache")
            if self.resources_changed_callback is not None:
                await self.resources_changed_callback()

    async def handle_request(self, request: types.ServerRequest):
        """handle a request"""
        
        logger.debug(f"Handling request: {request}")

        # handle ping
        if isinstance(request.root, types.PingRequest):
            return {}

        # handle sampling
        elif isinstance(request.root, types.CreateMessageRequest):
            if self.sampling_callback is None:
                logger.error("Sampling callback not set but received sampling request")
                return
            
            sampling_result = await self.sampling_callback(request.root)

            assert isinstance(sampling_result, types.CreateMessageResult), "Sampling callback must return a CreateMessageResult"
            
            return sampling_result.model_dump()

        # handle list roots
        elif isinstance(request.root, types.ListRootsRequest):
            if self.roots_callback is None:
                logger.error("Roots callback not set but received roots request")
                return

            roots_result = await self.roots_callback(request.root)

            assert isinstance(roots_result, types.ListRootsResult), "Roots callback must return a ListRootsResult"

            return roots_result.model_dump()

        # throw error for unknown request
        else:
            logger.error(f"Unknown request type: {request.root}")
            return



import asyncio
from easymcp.client.sessions.mcp import MCPClientSession
from easymcp.client.transports.docker import DockerTransport, DockerServerParameters


async def main():
    args = DockerServerParameters(image="mcp/time")
    transport = DockerTransport(args)

    client_session = MCPClientSession(transport)
    await client_session.init()

    result = await client_session.start()
    print(f"{result=}")

    # await asyncio.sleep(10)

    resources = await client_session.list_resources()
    print(f"{resources=}")

    tools = await client_session.list_tools()
    print(f"{tools=}")

    await client_session.stop()
    print("stopped")


asyncio.run(main())



from uuid import uuid4
from re import sub

from mcp.types import ClientRequest, JSONRPCRequest


def CreateJsonRPCRequest(request: ClientRequest) -> JSONRPCRequest:
    """Create a JSON RPC request"""
    return JSONRPCRequest(jsonrpc="2.0", id=str(uuid4()), **request.model_dump())

def format_server_name(server_name: str) -> str:
    """Format a server name to be namespacing friendly"""

    server_name = server_name.replace("_", "-")
    server_name = server_name.replace(".", "-")

    return sub(r'[^a-zA-Z0-9-]', '', server_name)


from pydantic import BaseModel
from easymcp.client.transports.generic import TransportProtocol
from mcp.client.sse import sse_client
from mcp import types

class SseServerParameters(BaseModel):
    """SseServerParameters class"""

    url: str
    """url to connect to"""

    headers: dict[str, str] = {}
    """headers to send"""

    timeout: int = 60
    """timeout in seconds"""


class SseTransport(TransportProtocol):
    """SseTransport class"""

    args: SseServerParameters
    connection = None

    readstream = None
    writestream = None

    def __init__(self, arguments: SseServerParameters):
        self.state = "constructed"
        self.args = arguments.model_copy(deep=True)

    async def init(self):
        """Perform init logic"""
        self.state = "initialized"
        self.client = sse_client(self.args.url, headers=self.args.headers, timeout=self.args.timeout)

    async def start(self):
        """Start the transport"""
        self.state = "started"

        self.readstream, self.writestream = await self.client.__aenter__()


    async def send(self, message: str):
        """Send data to the transport"""
        assert self.writestream, "Transport not started"

        msg = types.JSONRPCMessage.model_validate_json(message)
        await self.writestream.send(msg)

    async def receive(self) -> str:
        """Receive data from the transport"""
        assert self.readstream, "Transport not started"

        msg = await self.readstream.receive()
        
        if isinstance(msg, Exception):
            await self.stop()
            raise msg
        
        return msg.model_dump_json()

    async def stop(self):
        """Stop the transport"""
        self.state = "stopped"



import asyncio
import io
import os
import shutil
import sys
from typing import Literal

from loguru import logger
from pydantic import BaseModel

from easymcp.client.transports.generic import TransportProtocol


class StdioServerParameters(BaseModel):
    """Configuration for StdioTransport."""

    command: str
    """command to run"""

    args: list[str] = []
    """arguments to pass to the command"""

    env: dict[str, str] = {}
    """environment variables to set"""

    cwd: str = os.path.curdir
    """current working directory"""

    log_stderr: bool = True


class ReadBuffer:
    """Buffered reader using BytesIO to handle chunked stdio input."""

    def __init__(self):
        self.buffer = io.BytesIO()

    def append(self, data: bytes):
        self.buffer.seek(0, io.SEEK_END)
        self.buffer.write(data)

    def read_message(self) -> str | None:
        self.buffer.seek(0)
        data = self.buffer.getvalue()

        if b"\n" in data:
            message, rest = data.split(b"\n", 1)
            self.buffer = io.BytesIO(rest)
            return message.decode().strip()

        return None


class StdioTransport(TransportProtocol):
    """Asynchronous stdio transport."""

    state: Literal["constructed", "initialized", "started", "stopped"]
    arguments: StdioServerParameters
    subprocess: asyncio.subprocess.Process | None
    read_buffer: ReadBuffer

    def __init__(self, arguments: StdioServerParameters):
        self.state = "constructed"
        self.arguments = arguments.model_copy(deep=True)
        self.read_buffer = ReadBuffer()
        self.subprocess = None
        self.stderr_task: asyncio.Task | None = None

    async def init(self) -> None:
        self.state = "initialized"

        self.arguments.command = (
            shutil.which(self.arguments.command) or self.arguments.command
        )

        env = os.environ.copy()
        env.update(self.arguments.env)
        self.arguments.env = env

    async def start(self) -> None:
        self.state = "started"

        self.subprocess = await asyncio.create_subprocess_exec(
            self.arguments.command,
            *self.arguments.args,
            cwd=self.arguments.cwd,
            env=self.arguments.env,
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

        if self.arguments.log_stderr:
            self.stderr_task = asyncio.create_task(self.read_stderr())

    async def send(self, message: str) -> None:
        assert self.subprocess and self.subprocess.stdin, "subprocess stdin not open"
        formatted_message = message.strip() + "\n"
        logger.debug(f"Sending message: {formatted_message}")
        self.subprocess.stdin.write(formatted_message.encode())
        await self.subprocess.stdin.drain()

    async def receive(self) -> str:
        assert self.subprocess and self.subprocess.stdout, "subprocess stdout not open"

        while True:
            chunk = await self.subprocess.stdout.read(1024)
            if not chunk:
                break

            self.read_buffer.append(chunk)

            message = self.read_buffer.read_message()
            if message:
                logger.debug(f"Received message: {message}")
                return message

        raise RuntimeError("Subprocess stdout closed before complete message")

    async def read_stderr(self) -> None:
        if self.subprocess is None or self.subprocess.stderr is None:
            return

        async for line in self.subprocess.stderr:
            print(line.decode(), file=sys.stderr, end="")

    async def stop(self) -> None:
        try:
            if self.subprocess:
                self.subprocess.terminate()
                try:
                    await asyncio.wait_for(self.subprocess.wait(), timeout=5)
                except asyncio.TimeoutError:
                    self.subprocess.kill()
                    await self.subprocess.wait()

            logger.info("Transport stopped successfully.")
        except Exception as e:
            logger.error(f"Error stopping transport: {e}")
        finally:
            self.subprocess = None
            self.state = "stopped"



from inspect import iscoroutinefunction
from typing import Awaitable, Callable

from mcp import types
from pydantic import AnyUrl

from easymcp.client.SessionMaker import make_transport, transportTypes
from easymcp.client.sessions.GenericSession import (
    BaseSessionProtocol,
    LifeSpanProtocol,
    PromptsCompatible,
    PushingPromptsCompatible,
    PushingResourcesCompatible,
    PushingRootsCompatible,
    PushingSamplingCompatible,
    PushingToolsCompatible,
    ResourcesCompatible,
    ToolsCompatible,
)
from easymcp.client.utils import format_server_name


class ClientManager:
    """ClientManager class"""

    default_list_roots_callback: (
        Callable[[types.ListRootsRequest], Awaitable[types.ListRootsResult]] | None
    ) = None
    default_sampling_callback: (
        Callable[[types.CreateMessageRequest], Awaitable[types.CreateMessageResult]]
        | None
    ) = None

    default_list_tools_changed_callback: Callable[[], Awaitable[None]] | None = None
    default_list_prompts_changed_callback: Callable[[], Awaitable[None]] | None = None
    default_list_resources_changed_callback: Callable[[], Awaitable[None]] | None = None

    sessions: dict[str, BaseSessionProtocol] = dict()

    def __init__(self):
        pass

    async def init(self, servers: dict[str, transportTypes]):
        """initialize the client manager"""

        for server_name, server in servers.items():
            server_name = format_server_name(server_name)

            if server_name in self.sessions:
                raise ValueError(f"Server name {server_name} already exists")

            session = make_transport(server)
            self.sessions[server_name] = session

            await session.init()

            if self.default_list_roots_callback is not None:
                if isinstance(session, PushingRootsCompatible):
                    await session.register_roots_callback(
                        self.default_list_roots_callback
                    )

            if self.default_sampling_callback is not None:
                if isinstance(session, PushingSamplingCompatible):
                    await session.register_sampling_callback(
                        self.default_sampling_callback
                    )

            if isinstance(session, LifeSpanProtocol):
                await session.start()

    async def add_server(self, name: str, transport: transportTypes):
        """add a server to the manager"""

        name = format_server_name(name)

        if name in self.sessions:
            raise ValueError(f"Session {name} already exists")

        session = make_transport(transport)
        await session.init()

        if self.default_list_roots_callback is not None:
            if isinstance(session, PushingRootsCompatible):
                await session.register_roots_callback(self.default_list_roots_callback)

        if self.default_sampling_callback is not None:
            if isinstance(session, PushingSamplingCompatible):
                await session.register_sampling_callback(self.default_sampling_callback)

        if isinstance(session, LifeSpanProtocol):
            await session.start()

        self.sessions[name] = session

        return True

    async def remove_server(self, name: str):
        """remove a server from the manager"""

        name = format_server_name(name)

        if name not in self.sessions:
            raise ValueError(f"Session {name} does not exist")

        session = self.sessions[name]
        if isinstance(session, LifeSpanProtocol):
            await session.stop()
        del self.sessions[name]

        return True

    def list_servers(self):
        """list available servers"""

        return list(self.sessions.keys())

    async def list_tools(self, force: bool = False) -> list[types.Tool]:
        """list tools on all servers"""

        result: list[types.Tool] = []

        for name, session in self.sessions.items():
            if not isinstance(session, ToolsCompatible):
                continue

            tools = await session.list_tools(force=force)
            if tools is None:
                continue
            for tool in tools.tools:
                tool.name = f"{name}.{tool.name}"
                result.append(tool)

        return result

    async def call_tool(self, name: str, args: dict):
        """call a tool"""

        if "." not in name:
            raise ValueError("Tool name must be in the format <server>.<tool>")

        server_name, tool_name = name.split(".", 1)
        session = self.sessions.get(server_name)

        if session is None:
            raise ValueError(f"Server {server_name} not found")

        if not isinstance(session, ToolsCompatible):
            raise ValueError(f"Server {server_name} does not support tools")

        return await session.call_tool(tool_name, args)

    async def list_resources(self) -> list[types.Resource]:
        """list resources on all servers"""

        result: list[types.Resource] = []

        for name, session in self.sessions.items():
            if not isinstance(session, ResourcesCompatible):
                continue

            resources = await session.list_resources()
            if resources is None:
                continue
            for resource in resources.resources:
                # do not map known schemes to mcp
                if resource.uri.scheme not in (
                    "http",
                    "https",
                ):
                    resource.uri = AnyUrl(f"mcp-{name}+{resource.uri}")

                result.append(resource)

        return result

    async def read_resource(self, uri: AnyUrl | str):
        """read a resource"""

        if not isinstance(uri, AnyUrl):
            uri = AnyUrl(uri)

        if "+" not in uri.scheme:
            raise ValueError("Resource URI must be in the format mcp-<server>+<uri>")

        server_name, resource_scheme = uri.scheme.split("+", 1)
        server_name = server_name.removeprefix("mcp-")
        session = self.sessions.get(server_name)

        if session is None:
            raise ValueError(f"Server {server_name} not found")

        if not isinstance(session, ResourcesCompatible):
            raise ValueError(f"Server {server_name} does not support resources")

        # new_uri = str(URL(str(uri)).with_scheme(resource_scheme))
        new_uri = str(uri).removeprefix(f"mcp-{server_name}+")

        return await session.read_resource(new_uri)

    async def list_prompts(self) -> list[types.Prompt]:
        """list prompts on all servers"""

        result: list[types.Prompt] = []

        for name, session in self.sessions.items():
            if not isinstance(session, PromptsCompatible):
                continue

            prompts = await session.list_prompts()
            if prompts is None:
                continue
            for prompt in prompts.prompts:
                prompt.name = f"{name}.{prompt.name}"
                result.append(prompt)

        return result

    async def read_prompt(self, name: str, args: dict):
        """read a prompt"""
        raise NotImplementedError

    # callbacks
    async def register_roots_callback(
        self,
        callback: Callable[[types.ListRootsRequest], Awaitable[types.ListRootsResult]],
    ):
        """register a callback for roots"""
        assert callable(callback), f"{callback} must be callable"
        assert iscoroutinefunction(callback), f"{callback} must be an async function"

        self.default_list_roots_callback = callback

        sessions = filter(
            lambda x: isinstance(x, PushingRootsCompatible), self.sessions.values()
        )
        for session in sessions:
            assert isinstance(session, PushingRootsCompatible)  # makes mypy happy
            await session.register_roots_callback(callback)

    async def register_sampling_callback(
        self,
        callback: Callable[
            [types.CreateMessageRequest], Awaitable[types.CreateMessageResult]
        ],
    ):
        """register a callback for sampling"""
        assert callable(callback), f"{callback} must be callable"
        assert iscoroutinefunction(callback), f"{callback} must be an async function"

        self.default_sampling_callback = callback

        sessions = filter(
            lambda x: isinstance(x, PushingSamplingCompatible), self.sessions.values()
        )
        for session in sessions:
            assert isinstance(session, PushingSamplingCompatible)  # makes mypy happy
            await session.register_sampling_callback(callback)

    async def register_tools_changed_callback(
        self, callback: Callable[[], Awaitable[None]]
    ):
        """register a callback for tools changed"""
        assert callable(callback), f"{callback} must be callable"
        assert iscoroutinefunction(callback), f"{callback} must be an async function"

        self.default_list_tools_changed_callback = callback

        sessions = filter(
            lambda x: isinstance(x, PushingToolsCompatible), self.sessions.values()
        )
        for session in sessions:
            assert isinstance(session, PushingToolsCompatible)  # makes mypy happy
            await session.register_tools_changed_callback(callback)

    async def register_prompts_changed_callback(
        self, callback: Callable[[], Awaitable[None]]
    ):
        """register a callback for prompts changed"""
        assert callable(callback), f"{callback} must be callable"
        assert iscoroutinefunction(callback), f"{callback} must be an async function"

        self.default_list_prompts_changed_callback = callback

        sessions = filter(
            lambda x: isinstance(x, PushingPromptsCompatible), self.sessions.values()
        )
        for session in sessions:
            assert isinstance(session, PushingPromptsCompatible)  # makes mypy happy
            await session.register_prompts_changed_callback(callback)

    async def register_resources_changed_callback(
        self, callback: Callable[[], Awaitable[None]]
    ):
        """register a callback for resources changed"""
        assert callable(callback), f"{callback} must be callable"
        assert iscoroutinefunction(callback), f"{callback} must be an async function"

        self.default_list_resources_changed_callback = callback

        sessions = filter(
            lambda x: isinstance(x, PushingResourcesCompatible), self.sessions.values()
        )
        for session in sessions:
            assert isinstance(session, PushingResourcesCompatible)
            await session.register_resources_changed_callback(callback)



import asyncio
from easymcp.client.sessions.mcp import MCPClientSession
from easymcp.client.transports.sse import SseTransport, SseServerParameters


async def main():
    args = SseServerParameters(url="http://localhost:8000/sse")
    transport = SseTransport(args)

    client_session = MCPClientSession(transport)
    await client_session.init()

    result = await client_session.start()
    print(f"{result=}")

    # await asyncio.sleep(10)

    resources = await client_session.list_resources()
    print(f"{resources=}")

    tools = await client_session.list_tools()
    print(f"{tools=}")

    await asyncio.Future()


asyncio.run(main())



from asyncio import Queue, Task

from loguru import logger
import pydantic
from easymcp.client.transports.generic import TransportProtocol
from mcp import types


async def reader(transport: TransportProtocol, queue: Queue[types.JSONRPCMessage]):
    """Read data from the transport and put it in the queue"""

    async def _reader():
        while transport.state == "started":
            data = await transport.receive()

            try:
                parsed = types.JSONRPCMessage.model_validate_json(data)
            except pydantic.ValidationError:
                logger.error(f"Error parsing JSON: {data}")
                parsed = None

            if parsed is None:
                continue

            queue.put_nowait(parsed)

    task = Task(_reader())
    return task


async def writer(transport: TransportProtocol, queue: Queue[types.JSONRPCMessage]):
    """Write data from the queue to the transport"""

    async def _writer():
        while transport.state == "started":
            data = await queue.get()
            await transport.send(data.model_dump_json())

    task = Task(_writer())
    return task



import asyncio
from easymcp.client.transports.stdio import StdioTransport, StdioServerParameters


async def main():
    args = StdioServerParameters(command="echo", args=["Hello, world!"])
    transport = StdioTransport(args)
    await transport.init()
    await transport.start()
    print(f"{await transport.receive()=}")
    await transport.stop()


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
from easymcp.client.sessions.mcp import MCPClientSession
from easymcp.client.transports.stdio import StdioTransport, StdioServerParameters


async def main():
    args = StdioServerParameters(command="uvx", args=["mcp-timeserver"])
    transport = StdioTransport(args)

    client_session = MCPClientSession(transport)
    await client_session.init()

    result = await client_session.start()
    print(f"{result=}")

    # await asyncio.sleep(10)

    resources = await client_session.list_resources()
    print(f"{resources=}")

    tools = await client_session.list_tools()
    print(f"{tools=}")

    call = await client_session.call_tool("get-current-time", {})
    print(f"{call=}")

    resource = await client_session.read_resource("datetime://Asia/Chongqing/now")
    print(f"{resource=}")

    # await client_session.stop()
    await asyncio.Future()


asyncio.run(main())



from typing import Protocol, runtime_checkable, Literal
from pydantic import BaseModel


@runtime_checkable
class TransportProtocol(Protocol):
    """Generic transport interface"""

    state: Literal["constructed", "initialized", "started", "stopped"]

    def __init__(self, arguments: BaseModel) -> None: ...

    async def init(self) -> None:
        """perform init logic (e.g. pull docker containers, download code, etc.)"""
        ...

    async def start(self) -> None:
        """start the transport (e.g. spawn subprocesses, start docker containers, etc.)"""
        ...

    async def stop(self) -> None:
        """stop the transport"""
        ...

    async def send(self, message: str) -> None:
        """send data to the transport"""
        ...

    async def receive(self) -> str:
        """receive data from the transport"""
        ...
